@@ -115,6 +115,14 @@ int add(int x, int y) {
 	return x+y;
 };
 
+// escape vs noescape
+
+// TODO(#56378): enable in Go 1.23:
+// #cgo noescape handleGoStringPointerNoescape
+void handleGoStringPointerNoescape(void *s) {}
+
+void handleGoStringPointerEscape(void *s) {}
+
 // Following mimics vulkan complex definitions for benchmarking cgocheck overhead.
 
 typedef uint32_t VkFlags;
@@ -1106,6 +1114,18 @@ func benchCgoCall(b *testing.B) {
 			C.handleComplexPointer(&a0)
 		}
 	})
+	b.Run("string-pointer-escape", func(b *testing.B) {
+		for i := 0; i < b.N; i++ {
+			var s string
+			C.handleGoStringPointerEscape(unsafe.Pointer(&s))
+		}
+	})
+	b.Run("string-pointer-noescape", func(b *testing.B) {
+		for i := 0; i < b.N; i++ {
+			var s string
+			C.handleGoStringPointerNoescape(unsafe.Pointer(&s))
+		}
+	})
 	b.Run("eight-pointers", func(b *testing.B) {
 		var a0, a1, a2, a3, a4, a5, a6, a7 C.VkDeviceCreateInfo
 		for i := 0; i < b.N; i++ {
