@@ -59,11 +59,11 @@ func adddynrel(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loade
 			su.SetRelocSym(rIdx, syms.PLT)
 			su.SetRelocAdd(rIdx, r.Add()+int64(ldr.SymPlt(targ)))
 		}
-		if (targType == 0 || targType == sym.SXREF) && !ldr.AttrVisibilityHidden(targ) {
+		if targType == 0 || targType == sym.SXREF {
 			ldr.Errorf(s, "unknown symbol %s in RISCV call", ldr.SymName(targ))
 		}
 		su := ldr.MakeSymbolUpdater(s)
-		su.SetRelocType(rIdx, objabi.R_RISCV_PCREL_ITYPE)
+		su.SetRelocType(rIdx, objabi.R_RISCV_CALL)
 		return true
 
 	case objabi.ElfRelocOffset + objabi.RelocType(elf.R_RISCV_GOT_HI20):
@@ -99,11 +99,6 @@ func adddynrel(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loade
 		su.SetRelocType(rIdx, objabi.R_RISCV_PCREL_LO12_S)
 		return true
 
-	case objabi.ElfRelocOffset + objabi.RelocType(elf.R_RISCV_32_PCREL):
-		su := ldr.MakeSymbolUpdater(s)
-		su.SetRelocType(rIdx, objabi.R_RISCV_PCREL_32)
-		return true
-
 	case objabi.ElfRelocOffset + objabi.RelocType(elf.R_RISCV_RVC_BRANCH):
 		su := ldr.MakeSymbolUpdater(s)
 		su.SetRelocType(rIdx, objabi.R_RISCV_RVC_BRANCH)
@@ -135,7 +130,7 @@ func adddynrel(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loade
 	r = relocs.At(rIdx)
 
 	switch r.Type() {
-	case objabi.R_RISCV_PCREL_ITYPE:
+	case objabi.R_RISCV_CALL:
 		if targType != sym.SDYNIMPORT {
 			// nothing to do, the relocation will be laid out in reloc
 			return true
@@ -233,12 +228,12 @@ func elfreloc1(ctxt *ld.Link, out *ld.OutBuf, ldr *loader.Loader, s loader.Sym,
 		}
 		out.Write64(uint64(r.Xadd))
 
-	case objabi.R_RISCV_CALL, objabi.R_RISCV_CALL_TRAMP:
+	case objabi.R_RISCV_JAL, objabi.R_RISCV_JAL_TRAMP:
 		out.Write64(uint64(sectoff))
 		out.Write64(uint64(elf.R_RISCV_JAL) | uint64(elfsym)<<32)
 		out.Write64(uint64(r.Xadd))
 
-	case objabi.R_RISCV_PCREL_ITYPE, objabi.R_RISCV_PCREL_STYPE, objabi.R_RISCV_TLS_IE:
+	case objabi.R_RISCV_CALL, objabi.R_RISCV_PCREL_ITYPE, objabi.R_RISCV_PCREL_STYPE, objabi.R_RISCV_TLS_IE:
 		// Find the text symbol for the AUIPC instruction targeted
 		// by this relocation.
 		relocs := ldr.Relocs(s)
@@ -254,13 +249,14 @@ func elfreloc1(ctxt *ld.Link, out *ld.OutBuf, ldr *loader.Loader, s loader.Sym,
 		// corresponding R_RISCV_PCREL_LO12_I or R_RISCV_PCREL_LO12_S relocation.
 		// Note that the LO12 relocation must point to a target that has a valid
 		// HI20 PC-relative relocation text symbol, which in turn points to the
-		// given symbol. For further details see the ELF specification for RISC-V:
+		// given symbol. For further details see section 8.4.9 of the RISC-V ABIs
+		// Specification:
 		//
-		//   https://github.com/riscv-non-isa/riscv-elf-psabi-doc/blob/master/riscv-elf.adoc#pc-relative-symbol-addresses
+		//  https://github.com/riscv-non-isa/riscv-elf-psabi-doc/releases/download/v1.0/riscv-abi.pdf
 		//
 		var hiRel, loRel elf.R_RISCV
 		switch r.Type {
-		case objabi.R_RISCV_PCREL_ITYPE:
+		case objabi.R_RISCV_CALL, objabi.R_RISCV_PCREL_ITYPE:
 			hiRel, loRel = elf.R_RISCV_PCREL_HI20, elf.R_RISCV_PCREL_LO12_I
 		case objabi.R_RISCV_PCREL_STYPE:
 			hiRel, loRel = elf.R_RISCV_PCREL_HI20, elf.R_RISCV_PCREL_LO12_S
@@ -281,9 +277,6 @@ func elfreloc1(ctxt *ld.Link, out *ld.OutBuf, ldr *loader.Loader, s loader.Sym,
 		out.Write64(uint64(sectoff + 4))
 		out.Write64(uint64(elf.R_RISCV_TPREL_LO12_I) | uint64(elfsym)<<32)
 		out.Write64(uint64(r.Xadd))
-		out.Write64(uint64(sectoff + 8))
-		out.Write64(uint64(elf.R_RISCV_TPREL_ADD) | uint64(elfsym)<<32)
-		out.Write64(uint64(0))
 
 	default:
 		return false
@@ -300,7 +293,9 @@ func elfsetupplt(ctxt *ld.Link, ldr *loader.Loader, plt, gotplt *loader.SymbolBu
 		ctxt.Errorf(gotplt.Sym(), "got.plt is not empty")
 	}
 
-	// https://github.com/riscv/riscv-elf-psabi-doc/blob/master/riscv-elf.md#program-linkage-table
+	// See section 8.4.6 of the RISC-V ABIs Specification:
+	//
+	//  https://github.com/riscv-non-isa/riscv-elf-psabi-doc/releases/download/v1.0/riscv-abi.pdf
 	//
 	// 1:   auipc  t2, %pcrel_hi(.got.plt)
 	//      sub    t1, t1, t3               # shifted .got.plt offset + hdr size + 12
@@ -354,7 +349,9 @@ func addpltsym(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loade
 		panic("plt is not set up")
 	}
 
-	// https://github.com/riscv/riscv-elf-psabi-doc/blob/master/riscv-elf.md#program-linkage-table
+	// See section 8.4.6 of the RISC-V ABIs Specification:
+	//
+	//  https://github.com/riscv-non-isa/riscv-elf-psabi-doc/releases/download/v1.0/riscv-abi.pdf
 	//
 	// 1:  auipc   t3, %pcrel_hi(function@.got.plt)
 	//     l[w|d]  t3, %pcrel_lo(1b)(t3)
@@ -402,20 +399,20 @@ func archreloc(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, r loade
 	// If the call points to a trampoline, see if we can reach the symbol
 	// directly. This situation can occur when the relocation symbol is
 	// not assigned an address until after the trampolines are generated.
-	if r.Type() == objabi.R_RISCV_CALL_TRAMP {
+	if r.Type() == objabi.R_RISCV_JAL_TRAMP {
 		relocs := ldr.Relocs(rs)
 		if relocs.Count() != 1 {
 			ldr.Errorf(s, "trampoline %v has %d relocations", ldr.SymName(rs), relocs.Count())
 		}
 		tr := relocs.At(0)
-		if tr.Type() != objabi.R_RISCV_PCREL_ITYPE {
+		if tr.Type() != objabi.R_RISCV_CALL {
 			ldr.Errorf(s, "trampoline %v has unexpected relocation %v", ldr.SymName(rs), tr.Type())
 		}
 		trs := tr.Sym()
 		if ldr.SymValue(trs) != 0 && ldr.SymType(trs) != sym.SDYNIMPORT && ldr.SymType(trs) != sym.SUNDEFEXT {
 			trsOff := ldr.SymValue(trs) + tr.Add() - pc
 			if trsOff >= -(1<<20) && trsOff < (1<<20) {
-				r.SetType(objabi.R_RISCV_CALL)
+				r.SetType(objabi.R_RISCV_JAL)
 				r.SetSym(trs)
 				r.SetAdd(tr.Add())
 				rs = trs
@@ -426,14 +423,11 @@ func archreloc(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, r loade
 
 	if target.IsExternal() {
 		switch r.Type() {
-		case objabi.R_RISCV_CALL, objabi.R_RISCV_CALL_TRAMP:
+		case objabi.R_RISCV_JAL, objabi.R_RISCV_JAL_TRAMP:
 			return val, 1, true
 
-		case objabi.R_RISCV_PCREL_ITYPE, objabi.R_RISCV_PCREL_STYPE, objabi.R_RISCV_TLS_IE:
+		case objabi.R_RISCV_CALL, objabi.R_RISCV_PCREL_ITYPE, objabi.R_RISCV_PCREL_STYPE, objabi.R_RISCV_TLS_IE, objabi.R_RISCV_TLS_LE:
 			return val, 2, true
-
-		case objabi.R_RISCV_TLS_LE:
-			return val, 3, true
 		}
 
 		return val, 0, false
@@ -442,11 +436,11 @@ func archreloc(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, r loade
 	off := ldr.SymValue(rs) + r.Add() - pc
 
 	switch r.Type() {
-	case objabi.R_RISCV_CALL, objabi.R_RISCV_CALL_TRAMP:
+	case objabi.R_RISCV_JAL, objabi.R_RISCV_JAL_TRAMP:
 		// Generate instruction immediates.
 		imm, err := riscv.EncodeJImmediate(off)
 		if err != nil {
-			ldr.Errorf(s, "cannot encode R_RISCV_CALL relocation offset for %s: %v", ldr.SymName(rs), err)
+			ldr.Errorf(s, "cannot encode J-type instruction relocation offset for %s: %v", ldr.SymName(rs), err)
 		}
 		immMask := int64(riscv.JTypeImmMask)
 
@@ -576,35 +570,35 @@ func archreloc(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, r loade
 			panic(fmt.Sprintf("unknown relocation type: %v", r.Type()))
 		}
 
-		second := int64(uint32(val))
-		second = (second &^ immMask) | int64(uint32(imm))
-		return second, 0, true
+		ins := int64(uint32(val))
+		ins = (ins &^ immMask) | int64(uint32(imm))
+		return ins, 0, true
 
-	case objabi.R_RISCV_PCREL_ITYPE, objabi.R_RISCV_PCREL_STYPE:
+	case objabi.R_RISCV_CALL, objabi.R_RISCV_PCREL_ITYPE, objabi.R_RISCV_PCREL_STYPE:
 		// Generate AUIPC and second instruction immediates.
 		low, high, err := riscv.Split32BitImmediate(off)
 		if err != nil {
-			ldr.Errorf(s, "R_RISCV_PCREL_ relocation does not fit in 32 bits: %d", off)
+			ldr.Errorf(s, "pc-relative relocation does not fit in 32 bits: %d", off)
 		}
 
 		auipcImm, err := riscv.EncodeUImmediate(high)
 		if err != nil {
-			ldr.Errorf(s, "cannot encode R_RISCV_PCREL_ AUIPC relocation offset for %s: %v", ldr.SymName(rs), err)
+			ldr.Errorf(s, "cannot encode AUIPC relocation offset for %s: %v", ldr.SymName(rs), err)
 		}
 
 		var secondImm, secondImmMask int64
 		switch r.Type() {
-		case objabi.R_RISCV_PCREL_ITYPE:
+		case objabi.R_RISCV_CALL, objabi.R_RISCV_PCREL_ITYPE:
 			secondImmMask = riscv.ITypeImmMask
 			secondImm, err = riscv.EncodeIImmediate(low)
 			if err != nil {
-				ldr.Errorf(s, "cannot encode R_RISCV_PCREL_ITYPE I-type instruction relocation offset for %s: %v", ldr.SymName(rs), err)
+				ldr.Errorf(s, "cannot encode I-type instruction relocation offset for %s: %v", ldr.SymName(rs), err)
 			}
 		case objabi.R_RISCV_PCREL_STYPE:
 			secondImmMask = riscv.STypeImmMask
 			secondImm, err = riscv.EncodeSImmediate(low)
 			if err != nil {
-				ldr.Errorf(s, "cannot encode R_RISCV_PCREL_STYPE S-type instruction relocation offset for %s: %v", ldr.SymName(rs), err)
+				ldr.Errorf(s, "cannot encode S-type instruction relocation offset for %s: %v", ldr.SymName(rs), err)
 			}
 		default:
 			panic(fmt.Sprintf("unknown relocation type: %v", r.Type()))
@@ -617,11 +611,6 @@ func archreloc(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, r loade
 		second = (second &^ secondImmMask) | int64(uint32(secondImm))
 
 		return second<<32 | auipc, 0, true
-
-	case objabi.R_RISCV_PCREL_32:
-		pc := ldr.SymValue(s) + int64(r.Off())
-		off := ldr.SymValue(rs) + r.Add() - pc
-		return off, 0, true
 	}
 
 	return val, 0, false
@@ -634,10 +623,10 @@ func archrelocvariant(*ld.Target, *loader.Loader, loader.Reloc, sym.RelocVariant
 
 func extreloc(target *ld.Target, ldr *loader.Loader, r loader.Reloc, s loader.Sym) (loader.ExtReloc, bool) {
 	switch r.Type() {
-	case objabi.R_RISCV_CALL, objabi.R_RISCV_CALL_TRAMP:
+	case objabi.R_RISCV_JAL, objabi.R_RISCV_JAL_TRAMP:
 		return ld.ExtrelocSimple(ldr, r), true
 
-	case objabi.R_RISCV_PCREL_ITYPE, objabi.R_RISCV_PCREL_STYPE, objabi.R_RISCV_TLS_IE, objabi.R_RISCV_TLS_LE:
+	case objabi.R_RISCV_CALL, objabi.R_RISCV_PCREL_ITYPE, objabi.R_RISCV_PCREL_STYPE, objabi.R_RISCV_TLS_IE, objabi.R_RISCV_TLS_LE:
 		return ld.ExtrelocViaOuterSym(ldr, r, s), true
 	}
 	return loader.ExtReloc{}, false
@@ -648,7 +637,7 @@ func trampoline(ctxt *ld.Link, ldr *loader.Loader, ri int, rs, s loader.Sym) {
 	r := relocs.At(ri)
 
 	switch r.Type() {
-	case objabi.R_RISCV_CALL:
+	case objabi.R_RISCV_JAL:
 		pc := ldr.SymValue(s) + int64(r.Off())
 		off := ldr.SymValue(rs) + r.Add() - pc
 
@@ -702,13 +691,16 @@ func trampoline(ctxt *ld.Link, ldr *loader.Loader, ri int, rs, s loader.Sym) {
 			// address, so we have to assume a trampoline is required. Mark
 			// this as a call via a trampoline so that we can potentially
 			// switch to a direct call during relocation.
-			sb.SetRelocType(ri, objabi.R_RISCV_CALL_TRAMP)
+			sb.SetRelocType(ri, objabi.R_RISCV_JAL_TRAMP)
 		}
 		relocs := sb.Relocs()
 		r := relocs.At(ri)
 		r.SetSym(tramp)
 		r.SetAdd(0)
 
+	case objabi.R_RISCV_CALL:
+		// Nothing to do, already using AUIPC+JALR.
+
 	default:
 		ctxt.Errorf(s, "trampoline called with non-jump reloc: %d (%s)", r.Type(), sym.RelocName(ctxt.Arch, r.Type()))
 	}
@@ -718,7 +710,7 @@ func genCallTramp(arch *sys.Arch, linkmode ld.LinkMode, ldr *loader.Loader, tram
 	tramp.AddUint32(arch, 0x00000f97) // AUIPC	$0, X31
 	tramp.AddUint32(arch, 0x000f8067) // JALR	X0, (X31)
 
-	r, _ := tramp.AddRel(objabi.R_RISCV_PCREL_ITYPE)
+	r, _ := tramp.AddRel(objabi.R_RISCV_CALL)
 	r.SetSiz(8)
 	r.SetSym(target)
 	r.SetAdd(offset)
